/**
 * Compiz Peek
 *
 * Copyright (c) 2008 Sam Spilsbury <smspillaz@gmail.com>
 *
 * Based on opacify.c and thumbnail.c:
 *
 * Copyright (c) 2006 Kristian Lyngstøl <kristian@beryl-project.org>
 * Ported to Compiz and BCOP usage by Danny Baumann <maniac@beryl-project.org>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * Peek is a rip-off a Windows 7's peek feature, which makes windows
 * blocking the desktop transparent or the current taskbar window
 * the only visible window.
 * 
 * TODOs:
 * Make more options
 * borders
 * fades
 */

#include "peek.h"

COMPIZ_PLUGIN_20090315 (peek, PeekPluginVTable);

/* Core peek functions. These do the real work. ---------------------*/

/* Sets the real opacity and damages the window if actual opacity and
 * requested opacity differs. */
void
PeekWindow::setAttributes (int opacity, int brightness, int saturation)
{
    PEEK_SCREEN (screen);

    bool transparent;

    transparent = ((opacity != gWindow->paintAttrib ().opacity) ||
		   (brightness != gWindow->paintAttrib ().brightness) ||
		   (saturation != gWindow->paintAttrib ().saturation));

    if (mTransparent != transparent ||
        (gWindow->lastPaintAttrib ().opacity != opacity))
	cWindow->addDamage ();

    mOpacity = opacity;
    mBrightness = brightness;
    mSaturation = saturation;
    mTransparent = transparent;

    if (mTransparent)
	gWindow->glPaintSetEnabled (this, true);
    else if (ps->activeWindow != window)
	gWindow->glPaintSetEnabled (this, false);
	
}

bool
PeekWindow::checkPosition ()
{
    PEEK_SCREEN (screen);

    if (ps->optionGetCurrentViewport ())
    {
	int screenWidth = screen->width ();
	int screenHeight = screen->height ();
	if (window->serverX () >= screenWidth		       ||
	    window->serverX () + window->serverWidth () <= 0   ||
	    window->serverY () >= screenHeight		       ||
	    window->serverY () + window->serverHeight () <= 0)
	    return false;
    }

    return true;
}

CompWindow *
PeekScreen::findWindowAt (CompPoint p)
{
    foreach (CompWindow *cw, screen->windows ())
    {
	PEEK_WINDOW (cw);

	if (cw->iconGeometry ().isEmpty ())
	    continue;

	if (!cw->isMapped ())
	    continue;
	
	if (cw->state () & CompWindowStateSkipTaskbarMask)
	    continue;

	if (cw->state () & CompWindowStateSkipPagerMask)
	    continue;

	if (!cw->managed ())
	    continue;

	if (!pw->cWindow->pixmap ())
	    continue;

	if (cw->iconGeometry ().contains (p) &&
	    pw->checkPosition ())
	    return cw;
    }

    return NULL;
}

CompWindow *
PeekScreen::findDockAt (CompPoint p)
{
    CompWindow *found = NULL;

    foreach (CompWindow *dock, screen->windows ())
    {
	if (dock->type () & CompWindowTypeDockMask)
	{
	    if (dock->geometry ().contains (p))
	    {
		found = dock;
		break;
	    }
	}
    }

    return found;
}

void
PeekScreen::handleDockLeave ()
{
    foreach (CompWindow *w, screen->windows ())
    {
	PEEK_WINDOW (w);

    	if (pw->peekCandidate)
	    pw->setAttributes (pw->gWindow->paintAttrib ().opacity,
			       pw->gWindow->paintAttrib ().brightness,
			       pw->gWindow->paintAttrib ().saturation);
	pw->peekCandidate = false;
    }

    activeWindow = NULL;

    if (poller.active ())
	poller.stop ();

    gScreen->glPaintOutputSetEnabled (this, false);
}

void
PeekScreen::positionUpdate (const CompPoint &p)
{
    CompWindow *hovered, *dock;

    dock = findDockAt (p);
    if (!dock)
    {
	handleDockLeave ();
	return;
    }

    hovered = findWindowAt (p);
    if (hovered == activeWindow)
	return;

    activeWindow = hovered;

    foreach (CompWindow *w, screen->windows ())
    {
		PEEK_WINDOW (w);
	
		if (!hovered && optionGetDimNoneSelected () && p.x() > screen->width() - 20)
		//TODO: make the region an option
		{
		    pw->setAttributes (pw->gWindow->paintAttrib ().opacity,
				       pw->gWindow->paintAttrib ().brightness,
				       pw->gWindow->paintAttrib ().saturation);
		    pw->gWindow->glPaintSetEnabled (pw, false);
		    continue;
		}
	
		if (pw->peekCandidate)
		{
		    int opacity, brightness, saturation;
	
		    if (w == hovered)
		    {
			opacity = MIN (OPAQUE * optionGetActiveOpacity () / 100,
				      pw->gWindow->paintAttrib ().opacity);
			brightness = MIN (BRIGHT * optionGetActiveBrightness () / 100,
				      pw->gWindow->paintAttrib ().brightness);
			saturation = MIN (COLOR * optionGetActiveSaturation () / 100,
				      pw->gWindow->paintAttrib ().saturation);
		    }
		    else
		    {
			PEEK_WINDOW (w);
	
			opacity = MIN (OPAQUE * optionGetPassiveOpacity () / 100,
				      pw->gWindow->paintAttrib ().opacity);
			brightness = MIN (BRIGHT * optionGetPassiveBrightness () / 100,
				      pw->gWindow->paintAttrib ().brightness);
			saturation = MIN (COLOR * optionGetPassiveSaturation () / 100,
					  pw->gWindow->paintAttrib ().saturation);
	
			pw->gWindow->glPaintSetEnabled (pw, true);
		    }
	
		    pw->setAttributes (opacity, brightness, saturation);
	
		}
		else
		    pw->gWindow->glPaintSetEnabled (pw, false);
    }
}

void
PeekScreen::handleDockEnter ()
{
    foreach (CompWindow *w, screen->windows ())
    {
	PEEK_WINDOW (w);

	pw->peekCandidate = !w->shaded () && w->managed () &&
			    !(w->type () & CompWindowTypeDockMask) &&
			    optionGetWindowMatch ().evaluate (w);
	if (!poller.active())poller.start();
	CompPoint p=poller.getCurrentPosition ();
	printf("p.x() %d, screen->width %d\n",p.x(),screen->width());
	if (pw->peekCandidate && optionGetDimNoneSelected () && p.x() > screen->width() - 20)
	{
	    int opacity, brightness, saturation;

	    opacity = MIN (OPAQUE * optionGetPassiveOpacity () / 100,
			   pw->gWindow->paintAttrib ().opacity);
	    brightness = MIN (BRIGHT * optionGetPassiveBrightness () / 100,
			      pw->gWindow->paintAttrib ().brightness);
	    saturation = MIN (COLOR * optionGetPassiveSaturation () / 100,
			      pw->gWindow->paintAttrib ().saturation);

	    pw->setAttributes (opacity, brightness, saturation);
	}
    }

    //if (!poller.active ())poller.start ();

    gScreen->glPaintOutputSetEnabled (this, true);
}

bool
PeekScreen::handleTimeout (CompWindow *w)
{
    CompPoint  p;
    CompWindow *dock;

    p    = poller.getCurrentPosition ();
    dock = findDockAt (p);

    if (dock && dock == w)
    {
	handleDockEnter ();
	positionUpdate (p);
    }

    return false;
}

void
PeekWindow::paintBorder ()
{
#ifdef USE_BORDER_PAINT
    int wx = WIN_X (window);
    int wy = WIN_Y (window);
    int width = WIN_W (window);
    int height = WIN_H (window);
    int off = 5;

    PEEK_SCREEN (screen);

    glEnable (GL_BLEND);
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glColor4f (1.0, 1.0, 1.0, (float) pw->opacity / OPAQUE);

    foreach (CompTexture *tex, ps->windowTex)
    {
	tex.enable (Good);

	glBegin (GL_QUADS);

	glTexCoord2f (1, 0);
	glVertex2f (wx, wy);
	glVertex2f (wx, wy + height);
	glVertex2f (wx + width, wy + height);
	glVertex2f (wx + width, wy);

	glTexCoord2f (0, 1);
	glVertex2f (wx - off, wy - off);
	glTexCoord2f (0, 0);
	glVertex2f (wx - off, wy);
	glTexCoord2f (1, 0);
	glVertex2f (wx, wy);
	glTexCoord2f (1, 1);
	glVertex2f (wx, wy - off);

	glTexCoord2f (1, 1);
	glVertex2f (wx + width, wy - off);
	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy);
	glTexCoord2f (0, 0);
	glVertex2f (wx + width + off, wy);
	glTexCoord2f (0, 1);
	glVertex2f (wx + width + off, wy - off);

	glTexCoord2f (0, 0);
	glVertex2f (wx - off, wy + height);
	glTexCoord2f (0, 1);
	glVertex2f (wx - off, wy + height + off);
	glTexCoord2f (1, 1);
	glVertex2f (wx, wy + height + off);
	glTexCoord2f (1, 0);
	glVertex2f (wx, wy + height);

	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy + height);
	glTexCoord2f (1, 1);
	glVertex2f (wx + width, wy + height + off);
	glTexCoord2f (0, 1);
	glVertex2f (wx + width + off, wy + height + off);
	glTexCoord2f (0, 0);
	glVertex2f (wx + width + off, wy + height);

	glTexCoord2f (1, 1);
	glVertex2f (wx, wy - off);
	glTexCoord2f (1, 0);
	glVertex2f (wx, wy);
	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy);
	glTexCoord2f (1, 1);
	glVertex2f (wx + width, wy - off);

	glTexCoord2f (1, 0);
	glVertex2f (wx, wy + height);
	glTexCoord2f (1, 1);
	glVertex2f (wx, wy + height + off);
	glTexCoord2f (1, 1);
	glVertex2f (wx + width, wy + height + off);
	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy + height);

	glTexCoord2f (0, 0);
	glVertex2f (wx - off, wy);
	glTexCoord2f (0, 0);
	glVertex2f (wx - off, wy + height);
	glTexCoord2f (1, 0);
	glVertex2f (wx, wy + height);
	glTexCoord2f (1, 0);
	glVertex2f (wx, wy);

	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy);
	glTexCoord2f (1, 0);
	glVertex2f (wx + width, wy + height);
	glTexCoord2f (0, 0);
	glVertex2f (wx + width + off, wy + height);
	glTexCoord2f (0, 0);
	glVertex2f (wx + width + off, wy);

	glEnd ();
	tex.disable ();
    }
    glTexEnvf (GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_REPLACE);
    glDisable (GL_BLEND);
    glBlendFunc (GL_ONE, GL_ONE_MINUS_SRC_ALPHA);
#endif
}

bool
PeekScreen::glPaintOutput (const GLScreenPaintAttrib &attrib,
		           const GLMatrix            &transform,
		           const CompRegion          &region,
		           CompOutput                *output,
		           unsigned int              mask)
{
    bool status;

    status = gScreen->glPaintOutput (attrib, transform, region, output, mask);

    if (poller.active ())
    {
	GLMatrix sTransform (transform);

        sTransform.toScreenSpace (output, -DEFAULT_Z_CAMERA);

	glPushMatrix ();
	glLoadMatrixf (sTransform.getMatrix ());

#ifdef USE_BORDER_PAINT
	foreach (CompWindow *w, screen->windows ())
	{
	    PEEK_WINDOW (w);

	    if (pw->peekCandidate)
		pw->paintBorder ();
	}

#endif

	/* Don't paint the window twice */

	paintingWindowOnTop = true;

	/* Repaint the active window on top of all the bordered windows */
	if (activeWindow)
	{
	    PEEK_WINDOW (activeWindow);
	    pw->gWindow->glPaint (pw->gWindow->paintAttrib (),
			          transform, region, 0);
	}

	paintingWindowOnTop = false;

	glPopMatrix ();

    }

    return status;
}

bool
PeekWindow::glPaint (const GLWindowPaintAttrib &attrib,
		     const GLMatrix	       &transform,
		     const CompRegion	       &region,
		     unsigned int	       mask)
{
    if (mTransparent)
    {
	GLWindowPaintAttrib wAttrib = attrib;

	PEEK_SCREEN (screen);

	if (!((window == ps->activeWindow) && (ps->paintingWindowOnTop)))
	{
	    wAttrib.opacity = mOpacity;
	    wAttrib.brightness = mBrightness;
	    wAttrib.saturation = mSaturation;
	}
	else
	    mask |= PAINT_WINDOW_NO_CORE_INSTANCE_MASK;

	return gWindow->glPaint (wAttrib, transform, region, mask);
    }

    return gWindow->glPaint (attrib, transform, region, mask);
}

void
PeekScreen::handleEvent (XEvent *event)
{
    CompWindow *w;

    screen->handleEvent (event);

    switch (event->type) {
    case EnterNotify:
	if (displayTimeout.active ())
	    displayTimeout.stop ();

	w = screen->findWindow (event->xcrossing.window);
	if (w)
	{
	    if (w->type () & CompWindowTypeDockMask)
	    {
		displayTimeout.setCallback
			(boost::bind (&PeekScreen::handleTimeout, this, w));
		displayTimeout.start (optionGetTimeout (),
				      (float) optionGetTimeout () * 1.2);
	    }
	}
	break;
    default:
	break;
    }
}

/* Configuration, initialization, boring stuff. ----------------------- */

PeekWindow::PeekWindow (CompWindow *window) :
    PluginClassHandler <PeekWindow, CompWindow> (window),
    window (window),
    cWindow (CompositeWindow::get (window)),
    gWindow (GLWindow::get (window)),
    peekCandidate (false),
    mTransparent (false),
    mOpacity (100),
    mBrightness (100),
    mSaturation (100)
{
    GLWindowInterface::setHandler (gWindow, false);
}

PeekScreen::PeekScreen (CompScreen *screen) :
    PluginClassHandler <PeekScreen, CompScreen> (screen),
    cScreen (CompositeScreen::get (screen)),
    gScreen (GLScreen::get (screen)),
    activeWindow (NULL),
    paintingWindowOnTop (false)
{
    poller.setCallback (boost::bind (&PeekScreen::positionUpdate, this, _1));

    ScreenInterface::setHandler (screen);
    GLScreenInterface::setHandler (gScreen, false);
}

PeekScreen::~PeekScreen ()
{
    displayTimeout.stop ();
    poller.stop ();
}

bool
PeekPluginVTable::init ()
{
    if (!CompPlugin::checkPluginABI ("core", CORE_ABIVERSION))
	return false;

    if (!CompPlugin::checkPluginABI ("composite", COMPIZ_COMPOSITE_ABI))
	return false;

    if (!CompPlugin::checkPluginABI ("opengl", COMPIZ_OPENGL_ABI))
    {
	compLogMessage ("peek", CompLogLevelError,
			"No compatible rendering plugin found!");
	return false;
    }

    if (!CompPlugin::checkPluginABI ("mousepoll", COMPIZ_MOUSEPOLL_ABI))
    {
	compLogMessage ("peek", CompLogLevelError,
			"No compatible mousepoll plugin found!");
	return false;
    }

    return true;
}
