/**
 *
 * Compiz screensaver plugin
 *
 * screensaver.cpp
 *
 * Copyright (c) 2007 Nicolas Viennot <nicolas@viennot.biz>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 **/


#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "screensaver.h"
#include "rotatingcube.h"
#include "flyingwindows.h"

COMPIZ_PLUGIN_20090315 (screensaver, ScreenSaverPluginVTable);

template< typename _ScreenEffect, typename _WindowEffect >
static void screenSaverEffectInstance ()
{
	SCREENSAVER_SCREEN (screen);
	delete ss->mEffect;
	ss->mEffect = new _ScreenEffect ();
	
	foreach (CompWindow *w, screen->windows ())
	{
		SCREENSAVER_WINDOW(w);
		delete sw->mEffect;
		sw->mEffect = new _WindowEffect(w);
	}
}

void
ScreenSaverScreen::enableEffect ()
{
    enum ScreensaverOptions::Mode mode = (enum ScreensaverOptions::Mode) optionGetMode ();

    if(mode == ModeFlyingWindows)
	screenSaverEffectInstance<ScreenFlyingWindows, WindowFlyingWindows> ();

    if (mode == ScreensaverOptions::ModeRotatingCube)
	screenSaverEffectInstance <ScreenRotatingCube, WindowEffect> ();

    mTime = 0;

    if (!mEffect->enable ())
    {
	fprintf (stderr, "Fail\n");
	screenSaverEffectInstance <ScreenEffect, WindowEffect> ();
	return;
    }

    mState.fadingOut = false;
    mState.fadingIn = true;
    mState.running = true;
}

void
ScreenSaverScreen::disableEffect ()
{
    mEffect->disable ();
    mTime = 0;

    mState.fadingOut = true;
    mState.fadingIn = false;
}

void
ScreenSaverScreen::cleanEffect ()
{
    screenSaverEffectInstance <ScreenEffect, WindowEffect> ();
}

void
ScreenSaverScreen::setState (bool enable)
{
    if (!mState.running && enable)
	mEffect->loadEffect = true;
    if (mState.running && !enable && !mState.fadingOut)
	disableEffect ();
}

bool
ScreenSaverScreen::initiate (CompAction         *action,
			     CompAction::State  state,
			     CompOption::Vector options)
{

    setState (!mState.running);
    return true;
}

void
ScreenSaverScreen::cubeGetRotation (float &x, float &v, float &progress)
{
    mEffect->cubeGetRotation (x, v, progress);
}

bool
ScreenSaverWindow::glPaint (const GLWindowPaintAttrib &attrib,
			    const GLMatrix	      &transform,
			    const CompRegion	      &region,
			    unsigned int	      mask)
{
    return mEffect->glPaint (attrib, transform, region, mask);
}

bool
ScreenSaverScreen::glPaintOutput (const GLScreenPaintAttrib &attrib,
				  const GLMatrix	    &transform,
				  const CompRegion	    &region,
				  CompOutput		    *output,
				  unsigned int		    mask)
{
    return mEffect->glPaintOutput (attrib, transform, region, output, mask);
}

void
ScreenSaverScreen::preparePaint (int ms)
{
    mEffect->preparePaint (ms);
}

void
ScreenSaverScreen::glPaintTransformedOutput (const GLScreenPaintAttrib &attrib,
					     const GLMatrix	    &transform,
					     const CompRegion	    &region,
					     CompOutput		    *output,
					     unsigned int		    mask)
{
    mEffect->glPaintTransformedOutput (attrib, transform, region, output, mask);
}

void
ScreenSaverScreen::donePaint ()
{
    mEffect->donePaint ();
}

void
ScreenSaverScreen::handleEvent (XEvent *event)
{
    XScreenSaverNotifyEvent *xssEvent;

    mEffect->handleEvent (event);

    switch ((event->type & 0x7F) - mXSSContext.first_event)
    {
	case ScreenSaverNotify:
	    xssEvent = (XScreenSaverNotifyEvent *) event;
	    setState (xssEvent->state);
	    break;
	default:
	    break;
    }

    if (mEffect->loadEffect)
    {
	mEffect->loadEffect = false;
	enableEffect ();
    }
    else if (mEffect->cleanEffect)
    {
	mEffect->cleanEffect = false;
	cleanEffect ();
    }
}

void
ScreenSaverScreen::paint (CompOutput::ptrList &outputs,
			  unsigned int	      mask)
{
    mEffect->paint (outputs, mask);
}

void
ScreenSaverScreen::setXScreenSaver (bool enabled)
{
    Window root = screen->root ();

    if (enabled && !mXSSContext.init)
    {
	int dummy;
	long unsigned int mask = 0;
	XSetWindowAttributes attr;

	if (!XScreenSaverQueryExtension (screen->dpy (), &mXSSContext.first_event, &dummy))
	{
	    compLogMessage ("screensaver", CompLogLevelWarn,
			    "XScreenSaver Extension not available, cannot integrate with "\
			    "desktop settings!");
	    return;
	}

	mXSSContext.init = true;

	XGetScreenSaver (screen->dpy (), &mXSSContext.timeout, &mXSSContext.interval,
			 &mXSSContext.prefer_blanking, &mXSSContext.allow_exposures);
	XSetScreenSaver (screen->dpy (), (int) (optionGetAfter () * 60.0), mXSSContext.interval, 0, AllowExposures);

	XScreenSaverSetAttributes (screen->dpy (), root, -100, -100, 1, 1, 0,
				   CopyFromParent, CopyFromParent, DefaultVisual (screen->dpy (), screen->screenNum ()), mask, &attr);

	XScreenSaverSelectInput (screen->dpy (), root, ScreenSaverNotifyMask);

    }
    if (!enabled && mXSSContext.init )
    {
	mXSSContext.init = false;

	XSetScreenSaver (screen->dpy (), mXSSContext.timeout, mXSSContext.interval,
			 mXSSContext.prefer_blanking, mXSSContext.allow_exposures);

	XScreenSaverSelectInput (screen->dpy (), root, 0);
	XScreenSaverUnsetAttributes (screen->dpy (), root);
    }
}

void
ScreenSaverScreen::optionChanged (CompOption *opt,
				  ScreensaverOptions::Options num)
{
    setXScreenSaver (false);
    setXScreenSaver (optionGetStartAutomatically ());
}

ScreenSaverWindow::ScreenSaverWindow (CompWindow *w) :
    PluginClassHandler <ScreenSaverWindow, CompWindow> (w),
    mEffect (NULL),
    window (w),
    cWindow (CompositeWindow::get (w)),
    gWindow (GLWindow::get (w))
{
    ScreenSaverScreen *ss = ScreenSaverScreen::get (screen);

    if(ss->mState.running && (enum ScreensaverOptions::Mode) ss->optionGetMode() == ScreensaverOptions::ModeFlyingWindows)
	mEffect = new WindowFlyingWindows (w);
    else
	mEffect = new WindowEffect (w);


    GLWindowInterface::setHandler (gWindow);
}

ScreenSaverWindow::~ScreenSaverWindow ()
{
    delete mEffect;
}

ScreenSaverScreen::ScreenSaverScreen (CompScreen *s) :
    PluginClassHandler <ScreenSaverScreen, CompScreen> (s),
    cScreen (CompositeScreen::get (screen)),
    gScreen (GLScreen::get (screen)),
    cubeScreen (CubeScreen::get (screen)),
    mEffect (new ScreenEffect ()),
    mDesktopOpacity (OPAQUE)
{
    ScreenInterface::setHandler (screen);
    GLScreenInterface::setHandler (gScreen);
    CompositeScreenInterface::setHandler (cScreen);
    CubeScreenInterface::setHandler (cubeScreen);

    mState.running = false;
    mState.fadingOut = false;
    mState.fadingIn = false;

    optionSetInitiateKeyInitiate (boost::bind (&ScreenSaverScreen::initiate, this, _1, _2, _3));
    optionSetInitiateButtonInitiate (boost::bind (&ScreenSaverScreen::initiate, this, _1, _2, _3));
    optionSetInitiateEdgeInitiate (boost::bind (&ScreenSaverScreen::initiate, this, _1, _2, _3));

    optionSetStartAutomaticallyNotify (boost::bind (&ScreenSaverScreen::optionChanged, this, _1, _2));
    optionSetAfterNotify (boost::bind (&ScreenSaverScreen::optionChanged, this, _1, _2));

    mXSSContext.init = false;

    mEffect->loadEffect = false;
    mEffect->cleanEffect = false;

    setXScreenSaver (optionGetStartAutomatically ());
}

ScreenSaverScreen::~ScreenSaverScreen ()
{
    delete mEffect;
}


bool
ScreenSaverPluginVTable::init ()
{
    if (!CompPlugin::checkPluginABI ("core", CORE_ABIVERSION) ||
	!CompPlugin::checkPluginABI ("composite", COMPIZ_COMPOSITE_ABI) ||
	!CompPlugin::checkPluginABI ("opengl", COMPIZ_OPENGL_ABI) ||
	!CompPlugin::checkPluginABI ("cube", COMPIZ_CUBE_ABI))
	return false;

    return true;
}
