#ifndef SCREENSAVER_INTERNAL_H
#define SCREENSAVER_INTERNAL_H

#include <core/core.h>
#include <composite/composite.h>
#include <opengl/opengl.h>
#include <cube/cube.h>

#include <X11/extensions/scrnsaver.h>

#include "screensaver_options.h"
#include "matrix.h"
#include "vector.h"

#define WIN_X(w) ((w)->x () - (w)->input ().left )
#define WIN_Y(w) ((w)->y () - (w)->input ().top )
#define WIN_W(w) ((w)->width () + (w)->input ().left + (w)->input ().right )
#define WIN_H(w) ((w)->height () + (w)->input ().top + (w)->input ().bottom )

extern int displayPrivateIndex;

class ScreenSaverScreen;
class ScreenSaverWindow;

class ScreenWrapper
{
public:
	ScreenWrapper();
	~ScreenWrapper() {}

	virtual void cubeGetRotation( float& x, float& v, float &progress );
	virtual void handleEvent (XEvent *);
	virtual void preparePaint (int);
	virtual void donePaint ();
	virtual void glPaintTransformedOutput (const GLScreenPaintAttrib &,
					       const GLMatrix		 &,
					       const CompRegion		 &,
					       CompOutput		 *,
					       unsigned int		   );
	virtual bool glPaintOutput (const GLScreenPaintAttrib	 &,
			            const GLMatrix		 &,
			            const CompRegion		 &,
			            CompOutput		 	 *,
			            unsigned int		   );
	virtual void paint (CompOutput::ptrList &,
			    unsigned int	 );
						
protected:
	CompScreen* s;
	ScreenSaverScreen* ss;
};

class WindowWrapper
{
public:
	WindowWrapper( CompWindow* w );
	virtual ~WindowWrapper() {}
	virtual bool glPaint (const GLWindowPaintAttrib &,
			      const GLMatrix 	        &,
			      const CompRegion		&,
			      unsigned int		  );
protected:
	CompWindow* w;
	ScreenSaverWindow* sw;
};


class ScreenEffect : public ScreenWrapper
{
public:
	ScreenEffect() : ScreenWrapper(), progress(0)  {}
	virtual ~ScreenEffect() {}
	float getProgress() { return progress; }
	virtual void handleEvent (XEvent *event);
	
	virtual bool enable();
	virtual void disable() {}
	virtual void preparePaint( int msSinceLastPaint );
	virtual void cubeGetRotation (float &, float &, float &); // TODO: get rid of this!

	bool cleanEffect;
	bool loadEffect;
	
protected:
	virtual void clean() {}
	
private:
	float progress;
};

class WindowEffect : public WindowWrapper
{
public:
	WindowEffect( CompWindow* w ) : WindowWrapper(w) {}
	virtual ~WindowEffect() {}
};

class ScreenSaver
{
   public:

	class State
	{
	    public:
		// when the screensaver is enabled, running and fadingIn are set to true
		// then, after fadingInDuration, fadingIn is set to false
		// when the screensaver is requested to stop, fadingOut is set to true
		// after fadingOutDuration, running is set to false
		bool running;
		bool fadingOut;
		bool fadingIn;
	};

	class XSSContext
	{
	    public:
		int timeout, interval, prefer_blanking, allow_exposures;
		int first_event;
		bool init;
	};
};

class ScreenSaverScreen :
    public PluginClassHandler <ScreenSaverScreen, CompScreen>,
    public ScreenInterface,
    public CompositeScreenInterface,
    public GLScreenInterface,
    public CubeScreenInterface,
    public ScreensaverOptions
{
    public:

	ScreenSaverScreen (CompScreen *s);
	~ScreenSaverScreen ();

    public:

	CompositeScreen *cScreen;
	GLScreen	*gScreen;
	CubeScreen	*cubeScreen;

	ScreenSaver::State mState;
	ScreenSaver::XSSContext mXSSContext;

	ScreenEffect	   *mEffect;

	void
	cubeGetRotation (float &x, float &v, float &progress);

	void
	handleEvent (XEvent *);

	void
	preparePaint (int);

	void
	donePaint ();

	void
	glPaintTransformedOutput (const GLScreenPaintAttrib &,
				  const GLMatrix	    &,
				  const CompRegion	    &,
				  CompOutput		    *,
				  unsigned int		      );


	bool glPaintOutput (const GLScreenPaintAttrib	 &,
		            const GLMatrix		 &,
		            const CompRegion		 &,
		            CompOutput		 	 *,
		            unsigned int		   );

	void
	paint (CompOutput::ptrList &outputs,
	       unsigned int	   mask);

	void enableEffect ();

	void disableEffect ();

	void cleanEffect ();

	void setState (bool);

	bool initiate (CompAction *, CompAction::State, CompOption::Vector);

	void setXScreenSaver (bool);
	void optionChanged (CompOption *, ScreensaverOptions::Options);

	int mTime;
	float mCubeRotX, mCubeRotV, mCubeProgress, mZCamera;
	float mCubeRotXFadeOut, mCubeRotVFadeOut, mZCameraFadeOut;

	Point mScreenCenter;
	Matrix  mCamera, mCameraMat;

	float	  mAngleCam;
	GLushort  mDesktopOpacity;
};

class ScreenSaverWindow :
    public PluginClassHandler <ScreenSaverWindow, CompWindow>,
    public GLWindowInterface
{
    public:

	ScreenSaverWindow (CompWindow *w);
	~ScreenSaverWindow ();

	WindowEffect *mEffect;

	bool
	glPaint (const GLWindowPaintAttrib &,
		 const GLMatrix		   &,
		 const CompRegion	   &,
		 unsigned int		     );

	CompWindow *window;
	CompositeWindow *cWindow;
	GLWindow	*gWindow;

};

#define SCREENSAVER_SCREEN(s)						      \
    ScreenSaverScreen *ss = ScreenSaverScreen::get (s);

#define SCREENSAVER_WINDOW(w)						      \
    ScreenSaverWindow *sw = ScreenSaverWindow::get (w);

class ScreenSaverPluginVTable :
    public CompPlugin::VTableForScreenAndWindow <ScreenSaverScreen, ScreenSaverWindow>
{
    public:

	bool init ();
};

#endif
