/*
 * Compiz Static Windows plugin
 *
 * static.c
 *
 * Copyright : (C) 2008 by Mark Thomas
 * E-mail    : markbt@efaref.net
 *
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include "static.h"

COMPIZ_PLUGIN_20090315 (static, StaticPluginVTable);

bool
StaticWindow::isStatic ()
{
    return StaticScreen::get (screen)->optionGetWindowMatch ().evaluate (window);
}

void
StaticScreen::preparePaint (int         msSinceLastPaint)
{
    /* Initially we are painting all windows.
     * This may change if we switch to transformed mode.
     */
    mStaticMode = STATIC_ALL;
    
    cScreen->preparePaint (msSinceLastPaint);
}

void
StaticScreen::glPaintTransformedOutput (const GLScreenPaintAttrib &attrib,
					const GLMatrix		  &transform,
					const CompRegion	  &region,
					CompOutput		  *output,
					unsigned int		  mask)
{
    /* We are now painting in transformed mode.
     * Start painting only normal windows.
     */
    mStaticMode = STATIC_NORMAL;
    gScreen->glPaintTransformedOutput (attrib, transform, region, output, mask);
}

void
StaticScreen::glApplyTransform (const GLScreenPaintAttrib &attrib,
				CompOutput		  *output,
				GLMatrix		  *transform)
{
    if (mStaticMode == STATIC_STATIC)
    {
	//int oldNumber = gScreen->glApplyTransformGetCurrentIndex ();
	//gScreen->glApplyTransformSetCurrentIndex (MAXSHORT);
        gScreen->glApplyTransform (attrib, output, transform);
        //gScreen->glApplyTransformSetCurrentIndex (oldNumber);
    }
    else
    {
        gScreen->glApplyTransform (attrib, output, transform);
    }
}

bool
StaticScreen::glPaintOutput (const GLScreenPaintAttrib &attrib,
			     const GLMatrix	       &transform,
			     const CompRegion	       &region,
			     CompOutput		       *output,
			     unsigned int	       mask)
{
    bool status;
    status = gScreen->glPaintOutput (attrib, transform, region, output, mask);

    if (mStaticMode == STATIC_NORMAL) 
    {
        GLMatrix sTransform (transform);
        
        sTransform.toScreenSpace (output, -DEFAULT_Z_CAMERA);

        /* We switched to Transformed mode somewhere along the line.
         * Now paint the dock windows untransformed by bypassing
         * the virtual table.
         */
        //mask &= ~PAINT_SCREEN_CLEAR_MASK;

        mStaticMode = STATIC_STATIC;
        
        foreach (CompWindow *w, screen->windows ())
        {
	     STATIC_WINDOW (w);
	     
	     if (!sw->isStatic ())
		continue;

	     if (!output->contains (w->inputRect ()))
		continue;

	    GLMatrix wMatrix;
	    GLFragment::Attrib fragment (sw->gWindow->paintAttrib ());

	    wMatrix.toScreenSpace (&screen->currentOutputDev (), -DEFAULT_Z_CAMERA);
	    mask |= PAINT_WINDOW_TRANSFORMED_MASK;

	    int oldIndex = sw->gWindow->glAddGeometryGetCurrentIndex ();
	    glPushMatrix ();
	    glLoadMatrixf (wMatrix.getMatrix ());
	    sw->gWindow->glAddGeometrySetCurrentIndex (MAXSHORT);
	    status = sw->gWindow->glDraw (wMatrix, fragment, infiniteRegion, mask);
	    sw->gWindow->glAddGeometrySetCurrentIndex (oldIndex);
	    glPopMatrix ();

	    mStaticMode = STATIC_NORMAL;
	}
    }
    return status;
}


bool
StaticWindow::glPaint (const GLWindowPaintAttrib &attrib,
		       const GLMatrix		 &transform,
		       const CompRegion		 &region,
		       unsigned int		 mask)
{
    bool status = TRUE;
    STATIC_SCREEN (screen);
    
    if (ss->mStaticMode == STATIC_NORMAL && isStatic ())
	mask |= PAINT_WINDOW_NO_CORE_INSTANCE_MASK;
    glEnable(GL_BLEND);
    status = gWindow->glPaint (attrib, transform, region, mask);
    glDisable(GL_BLEND);
    return status;
}


StaticScreen::StaticScreen (CompScreen *screen) :
    PluginClassHandler <StaticScreen, CompScreen> (screen),
    cScreen (CompositeScreen::get (screen)),
    gScreen (GLScreen::get (screen)),
    mStaticMode (STATIC_ALL)
{
    CompositeScreenInterface::setHandler (cScreen);
    GLScreenInterface::setHandler (gScreen);
}

StaticScreen::~StaticScreen ()
{
}

StaticWindow::StaticWindow (CompWindow *window) :
    PluginClassHandler <StaticWindow, CompWindow> (window),
    window (window),
    cWindow (CompositeWindow::get (window)),
    gWindow (GLWindow::get (window))
{
    GLWindowInterface::setHandler (gWindow);
}

StaticWindow::~StaticWindow ()
{
}

bool
StaticPluginVTable::init ()
{
    if (!CompPlugin::checkPluginABI ("core", CORE_ABIVERSION) ||
        !CompPlugin::checkPluginABI ("composite", COMPIZ_COMPOSITE_ABI) ||
        !CompPlugin::checkPluginABI ("opengl", COMPIZ_OPENGL_ABI))
        return false;
    else
	return true;
}
